#define F_CPU 1000000UL

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

//defines
#define DispTime	1 /*what the counter should count to before changing what part of the display is on*/
//#define RNGon		PORTB |= 0b00000100
//#define RNGoff	PORTB &= 0b11111011
#define startADC	ADCSRA |= 0b01000000
#define	ADCreading	ADCSRA & 0b01000000
#define btnPress	(~PINB & 0b00000010)
#define randBit		(PINB & 0b00000001)
#define dispOn		PORTA = (displayVar & 0b01111111) | (PORTA & 0b10000000) /*copies the first 7 bits of displayVar in to PORTA, leaves the eigth bit of PORTA alone*/
#define dispH1On	PORTA = (displayVar & 0b01111000) | (PORTA & 0b10000000) /*only uses 4 of the 7 segments*/
#define dispH2On	PORTA = (displayVar & 0b00000111) | (PORTA & 0b10000000) /*uses the other 3 segments*/
//this does the linear feedback shift on pool without mixing in a radnom bit
#define cycle 		pool = (pool >> 1) | ((((pool >> 0) ^ (pool >> 3)) & 1) << 24)
//this does the linear feedback shift on pool and mixes in a radnom bit
#define mixInBit(bit)	pool = (pool >> 1) | (((bit ^ (pool >> 0) ^ (pool >> 3)) & 1) << 24)

/*Thanks to http://www.avrtutor.com for the
idea of programming the display this way*/
//for common cahtode:
//clears the first 7 bits of PORTA and leaves the last one alone
/*#define dispOff		PORTA &= 0b10000000
#define clrDisp		displayVar = 0b00000000
#define dispH1Off	PORTA &= 0b10000111
#define dispH2Off	PORTA &= 0b11111000
#define a		displayVar |= (1<<0)
#define b		displayVar |= (1<<1)
#define c		displayVar |= (1<<2)
#define d		displayVar |= (1<<3)
#define e		displayVar |= (1<<4)
#define f		displayVar |= (1<<5)
#define g		displayVar |= (1<<6)*/

//for common anode:
#define dispOff		PORTA |= 0b01111111
#define dispH1Off	PORTA |= 0b01111000
#define dispH2Off	PORTA |= 0b00000111
#define clrDisp		displayVar = 0b01111111
#define a		displayVar &= ~(1<<6)
#define b		displayVar &= ~(1<<5)
#define c		displayVar &= ~(1<<3)
#define d		displayVar &= ~(1<<0)
#define e		displayVar &= ~(1<<1)
#define f		displayVar &= ~(1<<4)
#define g		displayVar &= ~(1<<2)


//Typedefs
typedef unsigned char u8;
typedef unsigned int u16;
typedef unsigned long u32;

//prototypes
void Display(u8 digit);
u8 NewRandomNumber(u8 max);
void UpdateMax(u8 * max);
u8 VonNeumanBit(void);

//globals
u8 displayVar = 0xFF; //holds the output for the 7-segment display
u8 dispState = 0; //1=H1on, 0=H2on
/*pool will hold the data for the simulated shift register. No matter what we
start it as, it contents will coverge to around half one and half zeros given
time, but seeding it this way should speed up the process*/ 
u32 pool = 0b1010101010101010101010101;

int main(void)
{
  //setup
  u8 max = 16; //this value is out of range, so it will trigger UpdateMax
		//since it will be differ from the potientiometer reading
  DDRA = 0b01111111; // set port A pins 0-6 for output and 7 for input
  PORTA = 0b01111111; // set port A pins 0-6 high, and 7 for no pullup
  DDRB = 0b00000100; //set port B for input except pin 2
  PORTB = 0b11111011; //set pull-up resistors on port B and pin 2 low

  ADMUX = 0b00000111; //set Vcc as analog reference and pin A7 for input
  ADCSRA = 0b10000010; //enable ADC and set the clock prescalar to 4
  ADCSRB = 0b00010000; //left align ADC result

  OCR0A = 1; //set to 1 so that the interrupt will be called soon
  TCCR0A |= 0b00000010; //set to clear counter on match
  TCCR0B |= 0b00000101; //select 1024 clock prescalar
  TIMSK0 |= 1<<1; //enable compare A interrupt

  dispOff;
  _delay_ms(25); //give some time for everything to settle
  for(u8 i = 0; i < 75; i ++) //fill the pool by mixing in a bunch of bits
    mixInBit(VonNeumanBit());

  sei();
  //main loop
  while(1)
  {
    _delay_ms(25); //want new nums ~30 times/sec = 33ms, generation takes some
    UpdateMax(&max);
    if(!btnPress)
    {
      cli(); //turn interupts of so the display can be turned of and remain so
      dispOff;
      //RNGon;
      _delay_ms(5); //let things settle for a bit after the display is off
      Display(NewRandomNumber(max));
      //RNGoff;
      sei();
    }
  }
  return 0;
}

void Display(u8 digit)
{
  clrDisp;
  switch (digit)
  {
    case 0 : a; b; c; d; e; f; break;
    case 1 : b; c; break;
    case 2 : a; b; d; e; g; break;
    case 3 : a; b; c; d; g; break;
    case 4 : b; c; f; g; break;
    case 5 : a; c; d; f; g; break;
    case 6 : a; c; d; e; f; g; break;
    case 7 : a; b; c; break;
    case 8 : a; b; c; d; e; f; g; break;
    case 9 : a; b; c; f; g; break;
    case 10 : a; b; c; e; f; g; break;
    case 11 : c; d; e; f; g; break;
    case 12 : a; d; e; f; break;
    case 13 : b; c; d; e; g; break;
    case 14 : a; d; e; f; g; break;
    case 15 : a; e; f; g; break;
    default: break;
  }
}

u8 NewRandomNumber(u8 max){
  u8 num, mask, length;
  mask = 0;
  length = 0;
  
  //length is the number of binary digits of max
  //mask is that many ones
  //e.g. if max is 9 (1001 in binary) then length is 4
  //and mask is 1111 in binary (15 in decimal)
  while(mask < max)
  {
    mask = (mask << 1) | 1;
    length++;
  }
  
  //mix 25 new bits in to the pool
  for(u8 i = 0; i < 25; i++)
    mixInBit(VonNeumanBit());

  //now we generate num. To do this, we take length bits out of the pool.
  //if the bits are no good (say max is 1001 but we take 1011 out of the pool)
  //then we perform the linear feedback shift length times on the pool
  //and try again
  while(1)
  {
    num = pool & mask;
    if(num <= max)
      break;
    //if we need to go again, shift in an appropriate num of new bits
    for(u8 i = 0; i < length; i++)
      cycle;
  }
    
  return num;
}

/*This function will update max if the potentiomenter is moved. Every time it's
changed, it will display the result for ~3/4s but will still be sensitive to
further changes during this time.*/
void UpdateMax(u8 * max)
{
  u8 count = 0;
  u8 newMax = 0;
  
  do{
    startADC;
    while(ADCreading){}; //wait until the ADC is done reading
    newMax = ADCH >> 4; //keeps the 4 most significant bits
			//which will give us a number 0-15 as required
    if(newMax != *max)
    {
      *max = newMax;
      Display(newMax);
      count = 250;
    }
    if(count > 0)
    {
      count--;
      _delay_ms(3); //3ms * 250 + comp time ~=.75 s, which is a good duration
    }
  }while(count > 0);
}

u8 VonNeumanBit(void)
{
  u8 orand = 0;
  do{
    orand = randBit;
    _delay_us(100); //give the random bit generator time to make a new bit
  }while(orand == randBit);
  return orand;
}


/*This interupt turns the display on and off to saver power. The segments of
the display are divided in to two groups, and only one of them is turned on at
a time. When one is turned on, the other is turned off. This is done ~1024
clock cycles.*/
SIGNAL(SIG_OUTPUT_COMPARE0A)
{
  if(dispState) //display H1 is on
  {
    dispH1On;
    dispH2Off;
    OCR0A = DispTime;
    dispState = 0;
  }
  else //display H2 is on
  {
    dispH2On;
    dispH1Off;
    OCR0A = DispTime;
    dispState = 1;
  }
}